from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.functional import SimpleLazyObject
from django.utils.module_loading import import_string

from rest_framework import serializers

get_model = apps.get_model

def get_protocol_handler():
    return getattr(settings, 'TG_PUBSUB_PROTOCOL_HANDLER', 'tg_pubsub.protocol.RequestServerProtocol')


def get_protocol_handler_klass():
    return import_string(get_protocol_handler())


def get_pubsub_server_host():
    return getattr(settings, 'TG_PUBSUB_HOST', 'localhost')


def get_pubsub_server_port():
    return getattr(settings, 'TG_PUBSUB_PORT', 8090)


def get_hello_packets():
    """ Get all packets to send right after doing websocket handshake

        TG_PUBSUB_HELLO_PACKETS: List of import paths to callables that must return an instance of BaseMessage
    :return:
    """
    from .messages import BaseMessage

    packets = getattr(settings, 'TG_PUBSUB_HELLO_PACKETS', [])
    res = []

    assert isinstance(packets, (list, tuple))

    for path in packets:
        fn = import_string(path)

        assert callable(fn)

        instance = fn()
        assert isinstance(instance, BaseMessage)

        res.append(instance)

    return res


def get_extra_models():
    """ Get extra models to mark as listenable. This is useful if one needs to listen to changes
        on a model that is coming from external apps so subclassing variant is impossible.

        TG_PUBSUB_EXTRA_MODELS: list(config_path, ...)

    :rtype: dict
    """
    from .models import ListenableModelMixin, ModelListenConfig

    extra = getattr(settings, 'TG_PUBSUB_EXTRA_MODELS', [])

    res = {}

    assert isinstance(extra, (list, tuple))

    for config_path in extra:
        klass = import_string(config_path)

        if not issubclass(klass, ModelListenConfig):
            raise ImproperlyConfigured('extra_models: Config path %s is not a ModelListenConfig' % config_path)

        if klass.model_path in res:
            raise ImproperlyConfigured('extra_models: Model %s already listenable' % klass.model_path)

        instance = klass()

        if isinstance(instance.model, ListenableModelMixin):
            raise ImproperlyConfigured('extra_models: Model %s already listenable' % klass.model_path)

        res[instance.model_path] = instance

    return res


extra_models = SimpleLazyObject(get_extra_models)



import json

from rest_framework.utils import encoders

from . import pubsub

from .config import get_model, extra_models
from .exceptions import IgnoreMessageException


class BaseMessage(object):
    MESSAGE_IDENTIFIER = 'base'

    def __init__(self, channels, data):
        assert ':' not in self.MESSAGE_IDENTIFIER

        if not isinstance(channels, (list, tuple)):
            channels = [channels, ]

        self.channels = channels
        self.data = data

    def as_message(self):
        return ':'.join([
            self.MESSAGE_IDENTIFIER,
            json.dumps(self.data, cls=encoders.JSONEncoder),
        ])

    def publish(self):
        """ Publish message to redis queue
        """
        message = self.as_message()

        for channel in self.channels:
            if isinstance(channel, (list, tuple)):
                channel = ':'.join(channel)

            pubsub.publish(channel, message)

    @classmethod
    def prepare_for_send(cls, ws, data):
        return data


class ModelChanged(BaseMessage):
    MESSAGE_IDENTIFIER = 'model'

    def __init__(self, klass, action, instance):
        # We publish to `django`, `django:app_name-model_name`, `django:app_name-model_name:action`
        channels = ['django', '%s-%s' % (klass._meta.app_label, klass._meta.object_name), action]

        super().__init__([channels[0:k] for k in range(1, len(channels) + 1)], {
            'app': klass._meta.app_label,
            'model': klass._meta.object_name,
            'action': action,
            'pk': instance.pk,
        })

    @classmethod
    def prepare_for_send(cls, ws, data):
        from .models import ListenableModelMixin

        model_path = '%s.%s' % (data['app'], data['model'])
        model = get_model(data['app'], data['model'])
        inst = model.objects.get(pk=data['pk'])

        if issubclass(model, ListenableModelMixin):
            has_access = inst.has_access
            pubsub_serialize = inst.pubsub_serialize
            serializer = inst.get_serializer(inst)

        else:
            if model_path in extra_models:
                has_access = extra_models[model_path].has_access
                pubsub_serialize = extra_models[model_path].pubsub_serialize
                serializer = extra_models[model_path].get_serializer(inst)

            else:
                raise Exception('Model %s is not listenable' % model)

        if not has_access(inst, ws.user):
            raise IgnoreMessageException("Ignoring update on %s.%s:%s:%s, not authorized" % (data['app'], data['model'],
                                                                                             data['action'],data['pk']))

        return {
            'model': '%s.%s' % (data['app'], data['model']),
            'action': data['action'],
            'pk': data['pk'],
            'data': pubsub_serialize(inst, serializer),
        }


registry = {
    BaseMessage.MESSAGE_IDENTIFIER: BaseMessage,
    ModelChanged.MESSAGE_IDENTIFIER: ModelChanged,
}



def test_imports():
    """ Ensures all submodules are importable
    Acts as a simple smoketest.
    """
    import tg_pubsub

    from tg_pubsub import config
    from tg_pubsub import exceptions
    from tg_pubsub import messages
    from tg_pubsub import models
    from tg_pubsub import protocol
    from tg_pubsub import pubsub
    from tg_pubsub import worker

    from tg_pubsub.management.commands.pubsub_server import Command

    assert config.extra_models == {}
    assert config.extra_models == config.get_extra_models()



import logging

from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver

from rest_framework.serializers import ModelSerializer

from .config import extra_models, get_model
from .messages import ModelChanged

logger = logging.getLogger('tg_pubsub')


class ListenableBase(object):
    """ Base mixin that declares the api for listenables
    """
    serializer_class = None
    serializer_fields = ('pk', )

    def pubsub_get_model(self):
        return self.__class__

    def get_serializer_class(self):
        """ Return the class to use for the serializer.
        """
        if self.serializer_class is None:
            class ListenableSerializer(ModelSerializer):
                class Meta:
                    model = self.pubsub_get_model()

                    fields = self.serializer_fields

            return ListenableSerializer

        return self.serializer_class

    def get_serializer(self, *args, **kwargs):
        return self.get_serializer_class()(*args, **kwargs)

    @classmethod
    def has_access(cls, instance, user):
        """ Does the given user has access to the model instance. This can be used to only send model data to specific users.

        (Note: The user may be anonymous)

        :param instance: Model instance
        :param user:     User connected via pubsub
        :rtype: bool
        """
        return True

    @classmethod
    def should_notify(cls, instance, action):
        """ Should the given instance send out a change notification. This can be used to limit publishes to only specific instances.

        :param instance: Model instance
        :param action:   Action that triggered the notification (created, saved, deleted)
        :rtype: bool
        """
        return True

    @classmethod
    def pubsub_serialize(cls, instance, serializer):
        """ Serialize the given model instance before sending it to the users.

        :param instance: Model instance
        :param serializer: Drf serializer
        :return:
        """
        return serializer.to_representation(instance)


class ListenableModelMixin(ListenableBase):
    """ Mixin that will mark model as 'listenable'. Such models will send messages to Redis queue whenever they're updated.
    """
    pass


class ModelListenConfig(ListenableModelMixin):
    """ Special class that can be used to mark external app models as listenable.

        via: TG_PUBSUB_EXTRA_MODELS
    """

    model_path = None

    def __init__(self):
        super().__init__()

        if self.model_path is None:
            raise ImproperlyConfigured('ModelListenConfig invalid model path %s' % self.model_path)

        self.model = get_model(self.model_path)

    def pubsub_get_model(self):
        return self.model


def is_model_listenable(cls, action, instance):
    if not issubclass(cls, models.Model):
        return False

    if issubclass(cls, ListenableModelMixin):
        should_notify = cls.should_notify

    else:
        model_path = '%s.%s' % (cls._meta.app_label, cls._meta.object_name)

        if model_path in extra_models:
            should_notify = extra_models[model_path].should_notify

        else:
            return False

    if not should_notify(instance, action):
        return False

    return True


def model_changed(cls, action, instance):
    # This should never cause problems for the outside code, so wrap it all in try: except:, just to be sure.
    try:
        if not is_model_listenable(cls, action, instance):
            return

        logging.info("model_changed(): %s-%s:%s:%s", cls._meta.app_label, cls._meta.object_name, action, str(instance.id))

        ModelChanged(cls, action, instance).publish()

    except Exception as e:
        logging.warning("Exception in model_changed(%s, %s, %s: %s): %s", cls, action, type(instance), getattr(instance, 'id', '-'), e)


@receiver(post_save)
def model_post_save_handler(sender, instance, created, **kwargs):
    if created:
        model_changed(sender, 'created', instance)
    else:
        model_changed(sender, 'saved', instance)


@receiver(pre_delete)
def model_pre_delete_handler(sender, instance, **kwargs):
    model_changed(sender, 'deleted', instance)



import os


DEBUG = True

SECRET_KEY='hunter2'

SITE_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

MEDIA_ROOT = os.path.join(SITE_ROOT, 'media')
MEDIA_URL = '/media/'

STATIC_ROOT = os.path.join(SITE_ROOT, 'assets')
STATIC_URL = '/static/'
STATICFILES_DIRS = (
    os.path.join(SITE_ROOT, 'static'),
)


INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',

    'rest_framework',
    'tg_pubsub',
]



import asyncio
import json
import logging

import websockets

from django.utils.encoding import force_text
from rest_framework.utils import encoders

from .exceptions import InvalidMessageException, IgnoreMessageException

from . import pubsub

from .config import get_protocol_handler_klass, get_hello_packets
from .messages import registry


logger = logging.getLogger('tg_pubsub.server')


class HandlerProtocol(object):
    def __init__(self, ws, request):
        self.socket = ws
        self.request = request

        assert self.socket is not None

    @property
    def open(self):
        return self.socket.open

    @property
    def user(self):
        if self.request is None:
            from django.contrib.auth.models import AnonymousUser
            return AnonymousUser()

        return self.request.user

    @property
    def logging_key(self):
        return 'tg_pubsub.handler-%s' % (self.user.pk or 'none')

    @asyncio.coroutine
    def send(self, data):
        if isinstance(data, dict):
            data = json.dumps(data, cls=encoders.JSONEncoder)

        yield from self.socket.send(data)


class WebSocketHandler(object):
    def __init__(self, ws, request):
        super().__init__()

        self.ws = HandlerProtocol(ws, request)
        self.logger = logging.getLogger(self.ws.logging_key)

    @asyncio.coroutine
    def run(self):
        yield from self.send_hello()
        yield from self.send_on_change()

    @asyncio.coroutine
    def send_hello(self):
        packets = get_hello_packets()

        self.logger.debug("Sending %d hello packets", len(packets))

        for packet in packets:
            yield from self.ws.send(packet.prepare_for_send(self.ws, packet.data))

        self.logger.debug("Sent %d hello packets", len(packets))

    @asyncio.coroutine
    def send_on_change(self):
        # Create Redis connection, subscribe channels
        self.logger.debug("Entering main Redis loop")
        r = pubsub.create_redis_connection()
        p = r.pubsub()

        # Subscribe to django channel.
        p.subscribe('django')

        while self.ws.open:
            msg = p.get_message(ignore_subscribe_messages=True)
            if msg is None:
                # If there was no message, sleep for one second, and try again
                yield from asyncio.sleep(1.0)
                continue

            try:
                self.logger.debug("Got update from Redis: %s", msg)
                identifier, data = self.message_valid(msg)

            except InvalidMessageException:
                continue

            else:
                try:
                    yield from self.ws.send(identifier.prepare_for_send(self.ws, data))

                except IgnoreMessageException as e:
                    self.logger.debug('%s' % e)

    @classmethod
    def message_valid(cls, message):
        msg_data = force_text(message['data'])

        if not msg_data or ':' not in msg_data:
            raise InvalidMessageException()

        identifier, data = force_text(msg_data).split(':', 1)

        if not (identifier and data):
            raise InvalidMessageException()

        # Only accept valid identifiers
        if identifier not in registry.keys():
            raise InvalidMessageException()

        # Parse the message body
        try:
            data = json.loads(data)

        except:
            raise InvalidMessageException()

        return registry[identifier], data


@asyncio.coroutine
def client_handler(ws, path, request=None):
    handler = WebSocketHandler(ws, request)

    yield from handler.run()


def run_pubsub_server(host, port):
    logger.info("Starting pubsub server on %s:%d", host, port)

    start_server = websockets.serve(client_handler, host, port, klass=get_protocol_handler_klass())

    asyncio.get_event_loop().run_until_complete(start_server)
    asyncio.get_event_loop().run_forever()



import websockets

from importlib import import_module
from urllib.parse import parse_qs, urlparse

from django.conf import settings
from django.utils.functional import SimpleLazyObject


class FakeRequest(object):
    def __init__(self, path, session_key):
        from django.contrib.auth.middleware import get_user

        self.session_key = session_key

        self.path = path
        self.user = SimpleLazyObject(lambda: get_user(self))

    @property
    def session(self):
        if not hasattr(self, '_session'):
            engine = import_module(settings.SESSION_ENGINE)

            setattr(self, '_session', engine.SessionStore(self.session_key))

        return getattr(self, '_session')


class RequestServerProtocol(websockets.server.WebSocketServerProtocol):
    """ WebSocketServerProtocol that gives handler a request-like
        object which might contain the session/user if token was provided.
    """

    TOKEN_PARAM = 'token'

    def get_handler_kwargs(self, path, get_header):
        query_dict = parse_qs(urlparse(path).query)

        session_key = query_dict.get(getattr(settings, 'TG_PUBSUB_TOKEN_PARAM', self.TOKEN_PARAM), [''])[0]

        # Construct a request
        request = FakeRequest(path, session_key)

        return {
            'request': request,
        }


class SessionRequiredServerProtocol(RequestServerProtocol):
    """ WebSocketServerProtocol which only allows handshakes with a valid token
    """

    def get_handler_kwargs(self, path, get_header):
        handler_kwargs = super().get_handler_kwargs(path, get_header)

        # If no token provided, stop the handshake.
        if not handler_kwargs['request'].session_key:
            raise websockets.InvalidOrigin('Permission denied')

        return handler_kwargs


class AnyUserServerProtocol(SessionRequiredServerProtocol):
    """ WebSocketServerProtocol implementation that allows any users (that provide a token)
    """

    def has_permissions(self, request):
        return request is not None

    def get_handler_kwargs(self, path, get_header):
        handler_kwargs = super().get_handler_kwargs(path, get_header)

        if not self.has_permissions(handler_kwargs['request']):
            raise websockets.InvalidOrigin('Permission denied')

        return handler_kwargs


class AnonymousUserServerProtocol(AnyUserServerProtocol):
    """ WebSocketServerProtocol implementation that only allows anonymous users
    """

    def has_permissions(self, request):
        if super().has_permissions(request):
            return request.user.is_anonymous()

        return False


class AuthenticatedUserServerProtocol(AnyUserServerProtocol):
    """ WebSocketServerProtocol implementation that only allows authenticated users
    """

    def has_permissions(self, request):
        if super().has_permissions(request):
            return request.user.is_authenticated()

        return False


class StaffUserServerProtocol(AuthenticatedUserServerProtocol):
    """ WebSocketServerProtocol implementation that only allows staff users
    """

    def has_permissions(self, request):
        if super().has_permissions(request):
            return request.user.is_staff() or request.user.is_superuser()

        return False


class SuperUserServerProtocol(StaffUserServerProtocol):
    """ WebSocketServerProtocol implementation that only allows superusers
    """

    def has_permissions(self, request):
        if super().has_permissions(request):
            return request.user.is_superuser()

        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'django>=1.8,<1.10',
    'djangorestframework>=3',
    'redis>=2.10.3',
    'websockets',
]

setup(
    name='tg-pubsub',
    version='0.1.1',
    description="Easy pubsub for django",
    long_description=readme + '\n\n' + history,
    author="Thorgate",
    author_email='code@thorgate.eu',
    url='https://github.com/thorgate/tg-pubsub',
    packages=[
        'tg_pubsub',
    ],
    package_dir={'tg_pubsub': 'tg_pubsub'},
    include_package_data=True,
    install_requires=requirements,
    dependency_links=[
        # We currently use our own fork, and also created a PR: https://github.com/aaugustin/websockets/pull/98
        'git+https://github.com/Jyrno42/websockets.git@18d26ea1d48ecadf6fc49d526a27814fb4b16362#egg=websockets-3.0',
    ],
    license="ISCL",
    zip_safe=False,
    keywords='tg-pubsub,tg_pubsub,pubsub,pub,sub,django,redis,queue',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



import logging

from django.conf import settings

import redis


logger = logging.getLogger('tg_pubsub')


def create_redis_connection():
    redis_host = getattr(settings, 'REDIS_HOST', 'localhost')
    redis_port = getattr(settings, 'REDIS_PORT', 6379)
    redis_db = getattr(settings, 'REDIS_DB', 0)

    # NB: for pubsub, the database number doesn't matter. At all.
    r = redis.StrictRedis(host=redis_host, port=redis_port, db=redis_db)
    try:
        r.ping()
    except redis.exceptions.ConnectionError as e:
        logger.warning("Redis isn't available, pubsub will be disabled: %s", e)
        return None

    logger.info("Connected to Redis")
    return r


redis_connection = create_redis_connection()


def publish(channel, message):
    if redis_connection is None:
        return

    try:
        redis_connection.publish(channel, message)
    except redis.exceptions.ConnectionError as e:
        logger.warning("Redis seems to have died: %s", e)



from django.core.management import BaseCommand

from ...worker import run_pubsub_server
from ...config import get_pubsub_server_host, get_pubsub_server_port


class Command(BaseCommand):
    help = "Runs websocket-based pubsub server"
    args = '[optional port number]'

    def handle(self, port='', *args, **options):
        port = int(port) if port else get_pubsub_server_port()
        run_pubsub_server(get_pubsub_server_host(), port)
